#!/usr/bin/perl

use strict;
require "drbllib.pl";

#################### Global Params
my $dig = "/usr/bin/dig";

#security limits
my $maxfilename = 128;
my $maxdnsname  = 128;
my $maxxmltext  = 256;
my $maxsources  = 100;  # max sources for work zone

#################### Globals

my @Queue=();  #zones to be processed
my %Zones;     #zone index for node access
my @Nodes=();  # DRBL nodes
my $debugLevel=10;

#################### MAIN
#init default values
my ($initMaster,$initZone,$initDir) = ('ns.sci-nnov.ru','vote.drbl.sandy.ru','report');
#my ($initMaster,$initZone,$initDir) = ('ns1.senecainfo.us','vote.drbl.senecainfo.us','report');

#process command line
if (@ARGV) {
  while($ARGV[0] =~ /^\-/){
    my $flg = shift @ARGV;
    $debugLevel++ if $flg eq '-d';
    die "Usage: drbl-explorer.pl master zone report_dir\n" if $flg eq '-h';
  }
  $initMaster = $ARGV[0] if $#ARGV>=0;
  $initZone = $ARGV[1] if $#ARGV>0;
  $initDir = $ARGV[2] if $#ARGV>1;
#  die "Usage: drbl-explorer.pl master zone report_dir\n" if 
}


# make node list
pullOut(name=>$initZone,
       master =>$initMaster);

#debug output
printHash(\@Nodes,"DRBL-Net") if $debugLevel>2;

# print nodes
makeMainReport($initDir);

#################### END

#################### Constructors

#create default node
sub createNode {
  my $newNode = {createdate=>getCurTime()};

  a2h($newNode,@_);
  $newNode->{name} = safeDNSName($newNode->{name});
  return $newNode;
}

sub createZone {
  my $newZone = {createdate=>getCurTime()};
  a2h($newZone,@_);
  $newZone->{name} = safeDNSName($newZone->{name});
  $newZone->{master} = safeDNSName($newZone->{master});

  return $newZone;
}

#sub createSource {
#  my $newSource = {
#		   zone => undef
#		   };
#  a2h($newSource,@_);
#  return $newSource;
#}
####################

# convert array of parameters to hash
sub a2h {
  my $hash = shift @_;
  my $v;
  while( my $key =shift @_ ) {
    $v = shift @_;
    $hash->{$key} = $v if defined $v;
  }
}

#### pullOut a DRBL net
sub pullOut {
  my($error);
#  a2h(\%Params,@_);
  push @Queue, &createZone(@_);
  while(not ($error = &processQueue)) {
    outNot("Zone processed");
  }
  outErr("Stop queue proccessing: $error");
  return $error;
}

# process single zone 
#
# zone parameters must be trusted
#
sub processQueue {
  my $zType = undef;
  my $nodeId;
  my $zone = shift @Queue;

  return "End of zone queue reached" unless $zone;
  return undef if exists $Zones{$zone->{name}};  # zone already processed

  #Ok it is a new zone let's get some thing about it

  # get header (DNS TXT RR)
  my $tmpSrc = suck_zone1("$dig \"\@$zone->{master}\" \"$zone->{name}\" TXT |","$zone->{name}.");
  $zType="Not accessable or empty" unless (keys(%$tmpSrc));

  my %zoneParams;
  extractDRBLParams(\%zoneParams,$tmpSrc);
  printHash(\%zoneParams,"Zone params") if $debugLevel>0;

  # try to determinate a type
  $zType = getType(\%zoneParams) unless defined $zType;

  # mark zone as processed
  $Zones{$zone->{name}} = undef;

  if ( $zType eq 'INFO') {
    $nodeId = processInfo($zone);
  } elsif ($zType eq 'WORK' or $zType eq 'VOTE' or $zType eq 'UNKNOWN') {
    $zone->{type} = $zType;
    outDeb("Zone $zType");
    if ( defined $zoneParams{'@'}->{info}->[0] ) {
      my ($z,$m,$zn);
      # create link to info zone
      if( parseTarget($zoneParams{'@'}->{info}->[0],\$zn,\$m) == 2) {
	$z=createZone(type => 'INFO',name=>$zn,master=>$m);
	printHash($z,"zone") if $debugLevel>0;
	$nodeId = processInfo($z);
	# check if information zone created succefuly
	if(defined $nodeId){
	  push @Nodes,$nodeId;
	} else {
	  outErr("Error in information zone");
	  push @Nodes,createNode(name => $zone->{name},
			     master => $zone->{master},
			     description => 'No description',
				     error=>"Error in informational zone");
	}
      } else {
	outErr("Bad info zone target: $zoneParams{'@'}->{info}->[0]");
	$nodeId = createNode(name => $zone->{name},
			     master => $zone->{master},
			     description => 'No description',
				     error=>"Bad link to informational zone");
      push @Nodes,$nodeId;
      }
    } else {
      outErr("Can't locate info zone target");
      $nodeId = createNode(name => $zone->{name},
			   master=> $zone->{master},
			   description =>undef,
			   type =>"erroneous",
			   error=>"No valid informational zone");
      push @Nodes,$nodeId;
    }
  } else {  # zone type not recognized
      outErr("Zone error.$zone->{name}: $zType");
      $nodeId = createNode(name => $zone->{name},
			   master=>$zone->{master}, 
			   description =>undef,
			   type =>"erroneous",
			   error=>"$zType");
      push @Nodes,$nodeId;
  }
  $Zones{$zone->{name}} = $nodeId unless defined $Zones{$zone->{name}};
  return undef;
}



################### Staff

# Read informational zone and create node structure
# parameter: zone structure
# return undef if zone cannot be processed
# info zone parameters must be trusted
sub processInfo {
  my($info) = @_;
  my($tmpSrc,%zoneParams);
  outDeb("Suck an info zone (AXFR)");
  $tmpSrc = suck_zone1("$dig \@$info->{master} $info->{name} AXFR |","$info->{name}.");
  extractDRBLParams(\%zoneParams,$tmpSrc);
  printHash(\%zoneParams,"Info") if $debugLevel>0;

  my($voteSet,$workSet,$nodeId)=([],[],undef);
  my($target,$zone,$master,$weight);

  my $pZone = $zoneParams{'@'};

  # get general node parameters
  if ( defined($pZone)) {
    $nodeId = createNode(
			 name => $info->{name},
			 description => $pZone->{desc}->[0],
			 policy => $pZone->{policy}->[0],
			 version => $pZone->{version}->[0],
			 zoneupdated => $pZone->{date}->[0]
			);
    return unless  $nodeId;
    $nodeId->{info}={name=>$info->{name},master=>$info->{master},type=>'INFO'};
  } else {
    outErr("No data in informational zone: $info->{name}");
    return;
  }

  #run through all i-zone sections
  while( my($zNick,$pZone) = each(%zoneParams)) {
    if ($pZone->{type}->[0]  eq 'VOTE') {
      #vote section
      if (parseTarget($pZone->{target}->[0],\$zone,\$master) == 2) {
	# pareseTarget returns trusted result
	push @$voteSet,createZone( name=>$zone,
				   master=>$master,
				   type=>'VOTE',
				   description=>$pZone->{desc}->[0],
				   weight=>$pZone->{weight}->[0]
				   );
      } else {
	push @$voteSet,createZone( name=>$zNick,error=>'Bad zone pointer');
      }
    } elsif ($pZone->{type}->[0]  eq 'WORK') {
      # work zone 
      if (parseTarget($pZone->{target}->[0],\$zone,\$master) == 2) {
	my $src = [];
	push @$workSet,createZone( name=>$zone,
				   master=>$master,
				   type=>'WORK',
				   description=>$pZone->{desc}->[0],
				   threshold=>$pZone->{threshold}->[0],
				   source => $src
				   );
	foreach (@{$pZone->{source}}) {
	  if (/^(Local:|\/|\.)/ ) { #local files
	    s/Local:\s*//s;
	    if ( /(.+)\/(\d+(\.\d+)?)/ ) {
	      push @$src,createZone(master=>'local',name=>$1,weight=>$2);
	    } else {
	      push @$src,createZone(master=>'local',name=>'Bad',error=>'Bad zone pointer');
	    }
	  } elsif( parseTarget($_,\$zone,\$master,\$weight)==3) {
	    push @$src,createZone(master=>$master,name=>$zone,weight=>$weight);
	    push @Queue,$$src[$#$src];
	    outDeb("add zone $zone in to queue");
	    printHash($Queue[$#Queue],'Zone added') if $debugLevel>0;
	  } else {
	    push @$src,createZone(master=>'local',name=>'Bad',error=>'Bad zone pointer');
	  }
	}
      } else {
	push @$voteSet,createZone( name=>$zNick,error=>'Bad zone pointer');
      }
    } 
  }
  if ($nodeId) {
    $nodeId->{work} = $workSet;
    $nodeId->{vote} = $voteSet;
    foreach (@$voteSet,@$workSet) {
      $Zones{$_->{name}}=$nodeId;
    }
  } else {
    $nodeId = createNode(name => $info->{name},error=>'Invalid zone');
  }
  printHash($nodeId,'New node') if $debugLevel>0;
  return $nodeId;
}
# Determinate zone type
#
# 
sub getType {
  my ($pHash) = @_;
  my $declType =  $pHash->{'@'}->{type}->[0];
  return $declType if $declType;
  return "UNKNOWN";
}
# setup hash pointed by $pHash
# using zone source pointed by $zone
#
# !!!!! RESULT NOT TRUSTED !!!!!!!!!! (perl security mean)
sub extractDRBLParams {
  my($pHash,$zone) = @_;;

  foreach my $host (keys %{$zone}) {
#    print "host: $host\n";
    if (defined $zone->{$host}->{'TXT'}) {
      my (@drblPairs) = split (';',join(';',@{$zone->{$host}->{'TXT'}}));
      foreach (@drblPairs) {
#	print "Param:$_\n";
	if (/^DRBL\-([\w]+)[:=]\s*(.+)\s*$/ ) {
#	  print"get $1 = $2\n";
	  if ( not defined $pHash -> {$host} ){
	    $pHash -> {$host} = { lc("$1") => []};
	  }
	  push @{$pHash->{$host}->{lc("$1")}},"$2";
	}
      }
    }
  }
#  print "keys: " .join('.',keys %{$pHash}) . "\n";
}

sub printHash {
  my($pHash,$prefix) = @_;

  my $t = ref($pHash);
  if ($t eq 'HASH') {
    while( my ($p,$v) = each(%$pHash)) {
      printHash($v,"$prefix->$p");
    }
  }elsif( $t eq 'ARRAY') {
    for(my $i=0 ; $i<=$#{$pHash} ; $i++){
      printHash($pHash->[$i],$prefix."[$i]");
    }
  } else {
    print "$prefix = <$pHash>\n";
  }
}

sub getCurTime {
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
  $mon += 1;
  $year += 1900;
  return "$mday.$mon.$year $hour:$min:$sec (MSK)";
}
################### Convertors ###########################


# safeBasic
# basic security checker
# parameters:
# in     - value to be checked
# $err   - pointer to scalar where security code will be placed
#          < 0 -fatal (intrusion excpression triggered) 0 - Ok, 0 > minor security
#          correction in result
# $mlen  - max allowed length
# return  - securited value or undef if intrusion
sub safeBasic {
  my ($in,$err,$mlen) = @_;
  my $out;
  $$err=0;
  $$err = -1 if $in =~ /\0/;
  $$err = -1 if length($in) > $mlen;
  return unless $$err == 0;
  return $out=$in;
}
sub safeDNSName {
  my ($in,$err) = @_;
  my $out;
  $out=safeBasic($in,$err,$maxdnsname);
  return unless $out;
  $out =~ s/\s//g;
  $out = lc($out);
  if (not $out =~ /^[a-z0-9\.\-]+$/) {
    $$err = -1;
    return;
  }
  return $out;
}
sub safeFile {
  my ($in,$err) = @_;
  my $out;
  $out=safeBasic($in,$err,$maxfilename);
  $out =~ s/[\s\|\&\\\/;]/_/g;
  return $out;
}
sub safeWeb {
  my ($in,$err) = @_;
  my $out=safeXML($in,$err);
  return $out;
}
sub safeXML {
  my ($in,$err) = @_;
  $in=safeBasic($in,$err,$maxxmltext);
  $in =~ s/&/&amp;/g;
  $in =~ s/</&lt;/g;
  $in =~ s/>/&gt;/g;
  $in =~ s/\'/&apos;/g;
  $in =~ s/\"/&quot;/g;
  return $in;
}
sub isFormatError {
  my($err) = @_;
  return if $err eq 0;
  die "Format string violation\n";
}


################### Reports ##############################

#  Parse @Nodes and store int in xml files
#  at $storePath directory
#

sub makeMainReport {
  my($storePath)=@_;
  my $node;
  my $schemePath="www.agk.nnov.ru/ns/drbl/0/0";
  my $xmlVersion='<?xml version="1.0" encoding="UTF-8"?>';
  my $netFile = "$storePath/net.xml";
  my $curTime = getCurTime();
  open netF, ">$netFile" or die "Can' open DRBL net index file for writing: $netFile: $!\n";

# write xml header
  print netF<<EOF;
$xmlVersion
<net xmlns="$schemePath" updated="$curTime">
EOF

  my ($nodeFile,$err,$XMLFName,$XMLName,$XMLPolicy,$XMLDescription,$XMLWorks,$XMLVotes);
  foreach $node (sort nodeCMP @Nodes) {
    my @indexBody=();
    my @nodeBody=();
    $nodeFile = safeFile($node->{name},\$err);
    isFormatError($err);
    $XMLFName=safeXML($nodeFile,\$err);
    isFormatError($err);
    $XMLName=safeXML($node->{name},\$err);
    isFormatError($err);
    $XMLPolicy = safeXML($node->{policy},\$err);
    isFormatError($err);
    $XMLDescription = safeXML($node->{description},\$err);
    isFormatError($err);

    $XMLWorks = formatWorks($node);
    $XMLVotes = formatVotes($node);
    push @nodeBody,"<description>$XMLDescription</description>" if defined $XMLDescription;
    push @nodeBody,"<warning>".&safeXML($node->{warning})."</warning>" if defined $node->{warning};
    push @nodeBody,"<version>".&safeXML($node->{version})."</version>" if defined $node->{version};
    push @nodeBody,"<updated>".&safeXML($node->{zoneupdated})."</version>" if defined $node->{zoneupdated};
    push @nodeBody,"<works>\n". $XMLWorks ."</works>"  if $XMLWorks;
    push @nodeBody,"<votes>\n". $XMLVotes ."</votes>"  if $XMLVotes;
    push @nodeBody,"<error>".&safeXML($node->{error})."</error>" if defined($node->{error}) and (@nodeBody);
    $nodeFile = "$storePath/$nodeFile";
    if (@nodeBody) {
      open nodeF,">$nodeFile.xml" or die  "Can' open DRBL node file for writing: $nodeFile: $!\n";
      print nodeF "$xmlVersion\n<?xml-stylesheet type=\"text/xsl\" href=\"nodeView.xsl\"?>\n";
      print nodeF "<node name=\"$XMLName\" xmlns=\"$schemePath\" updated=\"$node->{createdate}\"";
      print nodeF " policy=\"$XMLPolicy\"" if defined $XMLPolicy;
      print nodeF ">\n";   # close node open tag
      foreach (@nodeBody) {
	print nodeF "$_\n";
      }
      print nodeF "</node>\n"; # node close tag
      close nodeF;
    }
    #out node int net report
    print netF "<node name=\"$XMLName\" updated=\"$node->{createdate}\"";
    print netF " policy=\"$XMLPolicy\"" if defined $XMLPolicy;
    push @indexBody,"<description>$XMLDescription</description>" if defined $XMLDescription;
    push @indexBody,"<dlink ref=\"$XMLFName\">Node details</dlink>" if @nodeBody;
    push @indexBody,"<error>".&safeXML($node->{error})."</error>" if defined $node->{error};
    if (@indexBody) {
      print netF ">";
      foreach (@indexBody) {
	print netF "$_\n";
      }
      print netF "</node>\n";
    } else {
      print netF "/>";
    }
  }
print netF <<EOF;
</net>
EOF
  close netF;
}

sub formatVotes {
  my($node) = @_;
  my $str="";
  return unless defined  $node->{vote};
  foreach my $zone (@{$node->{vote}}) {
    my @body=();
    $str .="<vote ";
    push @body,'<description>'. &safeXML($zone->{description}).'</description>' if defined  $zone->{description};
    $str .= " master=\"". &safeXML($zone->{master}). "\"" if defined $zone->{master};
    $str .= " name=\"". &safeXML($zone->{name}). "\"" if defined $zone->{name};
    $str .= " weight=\"". &safeXML($zone->{weight}). "\"" if defined $zone->{weight};
    if (@body) {
      $str .= ">";
      foreach (@body){
	$str .= "$_\n";
      }
      $str .= "</vote>";
    } else {
      $str .= "/>";
    }
  }
  return $str;
}
sub formatWorks {
  my($node) = @_;
  my $str="";
  return unless defined  $node->{work};
  foreach my $zone (@{$node->{work}}) {
    my @body=();
    $str .="<work ";
    push @body,'<description>'. &safeXML($zone->{description}).'</description>' if defined  $zone->{description};
    $str .= " master=\"". &safeXML($zone->{master}). "\"" if defined $zone->{master};
    $str .= " name=\"". &safeXML($zone->{name}). "\"" if defined $zone->{name};
    $str .= " threshold=\"". &safeXML($zone->{threshold}). "\"" if defined $zone->{threshold};
    # defined sources
    foreach my $src (@{$zone->{source}}) {
      my $srcstr = '<source';
      $srcstr .= " master=\"". &safeXML($src->{master}). "\"" if defined $src->{master};
      $srcstr .= " name=\"". &safeXML($src->{name}). "\"" if defined $src->{name};
      $srcstr .= " weight=\"". &safeXML($src->{weight}). "\"" if defined $src->{weight};
      $srcstr .= "/>";
      push @body,$srcstr;
    }
    
    if (@body) {
      $str .= ">\n";
      foreach (@body){
	$str .= "$_\n";
      }
      $str .= "</work>";
    } else {
      $str .= "/>\n";
    }
  }
  return $str;
}

sub nodeCMP {
#  print "$a <> $b \n";
#  return $a <=> $b;
  return -1 if (not exists($a->{error})) 
    and   exists($b->{error});
  return 1 if (not exists($b->{error})) 
    and   exists($a->{error});
  return  - ($a->{name} <=> $b->{name});

}
